print("import: 'hypothesis'")
import hypothesis

print("import: 'hypothesis.extra'")
import hypothesis.extra

print("import: 'hypothesis.internal'")
import hypothesis.internal

print("import: 'hypothesis.internal.conjecture'")
import hypothesis.internal.conjecture

print("import: 'hypothesis.searchstrategy'")
import hypothesis.searchstrategy

print("import: 'hypothesis.tools'")
import hypothesis.tools

print("import: 'hypothesis.utils'")
import hypothesis.utils

print("import: 'hypothesis.vendor'")
import hypothesis.vendor

